package agent

import (
	"context"
	"errors"
	"fmt"

	"github.com/robfig/cron/v3"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/module/modagent"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/internal/module/starboard_vulnerability"
	"gitlab.com/gitlab-org/cluster-integration/gitlab-agent/v16/pkg/agentcfg"
	"go.uber.org/zap"
	"google.golang.org/protobuf/proto"
	"k8s.io/apimachinery/pkg/util/wait"
	"k8s.io/client-go/kubernetes"
)

var ErrMissingCadence = errors.New("cadence is a required field")

type module struct {
	log                       *zap.Logger
	api                       modagent.Api
	kubeClientset             kubernetes.Interface
	gitlabAgentNamespace      string
	gitlabAgentServiceAccount string
	holder                    *workerHolder
}

func (m *module) Run(ctx context.Context, repositoryConfig <-chan *agentcfg.AgentConfiguration) error {
	policiesUpdateData := make(chan *configurationToUpdateData)
	policiesHolder := m.newSecurityPoliciesWorkerHolder(ctx, policiesUpdateData)

	defer func() {
		policiesHolder.stop()
		if m.holder != nil {
			m.holder.stop()
		}
	}()

	var securityPoliciesEnabled bool

	for {
		select {
		case config, ok := <-repositoryConfig:
			if !ok {
				return nil
			}

			if !securityPoliciesEnabled {
				data := &configurationToUpdateData{
					agentId:                 config.AgentId,
					containerScanningConfig: config.ContainerScanning,
				}

				m.handleUpdate(ctx, data)
			}
		case data := <-policiesUpdateData:
			if data.containerScanningConfig == nil {
				m.log.Debug("ContainerScanning config is empty, security policies are disabled")
				securityPoliciesEnabled = false
			} else {
				m.log.Debug("ContainerScanning config is present, security policies are enabled")
				securityPoliciesEnabled = true
				m.handleUpdate(ctx, data)
			}
		}
	}
}

func (m *module) handleUpdate(ctx context.Context, data *configurationToUpdateData) {
	if m.holder != nil {
		if m.holder.configChanged(data.containerScanningConfig) {
			m.holder.stop()
			m.holder = m.newWorkerHolder(ctx, data)
		}
	} else {
		m.holder = m.newWorkerHolder(ctx, data)
	}
}

func (m *module) newSecurityPoliciesWorkerHolder(ctx context.Context, updater chan<- *configurationToUpdateData) *securityPoliciesWorkerHolder {
	workerHolder := &securityPoliciesWorkerHolder{
		worker: securityPoliciesWorker{
			api:     m.api,
			log:     m.log,
			updater: updater,
		},
	}
	ctx, workerHolder.cancel = context.WithCancel(ctx)
	workerHolder.wg.StartWithContext(ctx, workerHolder.worker.Run)

	return workerHolder

}

func (m *module) newWorkerHolder(ctx context.Context, cfg *configurationToUpdateData) *workerHolder {
	config := cfg.containerScanningConfig
	holder := &workerHolder{
		agentID: cfg.agentId,
		config:  cfg,
	}

	if config == nil {
		holder.cancel = func() {}
		return holder
	}

	cadence, _ := cron.ParseStandard(config.Cadence)

	var targetNamespaces []string
	if config.VulnerabilityReport != nil {
		targetNamespaces = config.VulnerabilityReport.Namespaces
	}

	scanJob := &scanJob{
		log:                       m.log,
		api:                       m.api,
		kubeClientset:             m.kubeClientset,
		gitlabAgentNamespace:      m.gitlabAgentNamespace,
		gitlabAgentServiceAccount: m.gitlabAgentServiceAccount,
		agentID:                   holder.agentID,
		targetNamespaces:          targetNamespaces,
		logParser:                 &logParserImpl{},
	}

	ctx, holder.cancel = context.WithCancel(ctx)
	w := &worker{
		job:     scanJob,
		cadence: cadence,
	}

	holder.wg.StartWithContext(ctx, w.Run)
	return holder
}

func (m *module) IsRunnableConfiguration(cfg *agentcfg.AgentConfiguration) bool {
	// starboard_vulnerability module is always enabled as it vulnerability scan could be enforced by security policies
	return true
}

func (m *module) DefaultAndValidateConfiguration(cfg *agentcfg.AgentConfiguration) error {
	if cfg.ContainerScanning == nil {
		return nil
	}

	if cfg.ContainerScanning.Cadence == "" {
		return ErrMissingCadence
	}

	if _, err := cron.ParseStandard(cfg.ContainerScanning.Cadence); err != nil {
		return fmt.Errorf("cadence is invalid: %w", err)
	}

	return nil
}

func (m *module) Name() string {
	return starboard_vulnerability.ModuleName
}

type workerHolder struct {
	agentID int64
	config  *configurationToUpdateData
	wg      wait.Group
	cancel  context.CancelFunc
}

type securityPoliciesWorkerHolder struct {
	worker securityPoliciesWorker
	wg     wait.Group
	cancel context.CancelFunc
}

func (h *securityPoliciesWorkerHolder) stop() {
	h.cancel()
	h.wg.Wait()
}

func (h *workerHolder) configChanged(containerScanningConfig *agentcfg.ContainerScanningCF) bool {
	return !proto.Equal(h.config.containerScanningConfig, containerScanningConfig)
}

func (h *workerHolder) stop() {
	h.cancel()
	h.wg.Wait()
}
